#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@author: David Siroky (siroky@dasir.cz)
"""

import types
import sys

import yaml
from lxml import etree
from lxml.builder import ElementMaker

#########################################################################
#########################################################################

PT_TO_CM = 1.0 / 72.0 * 2.54

DATA_LINE_WIDTH = 0.02 # cm
DATA_COLOR = "#000000"
DATA_FONT_SIZE = 9 # pt
ARROW_LENGTH = 0.2 # cm
ARROW_WIDTH = 0.15 # cm

MARK_COLOR = "#000000"
MARK_SIZE = 0.1 # cm

LABEL_CONNECTOR_COLOR = "#d0d0d0"
LABEL_CONNECTOR_WIDTH = 0.01 # cm

PROCESS_SPACING = 2.0 # cm
PROCESS_LINE_VOFFSET = 1.0 # cm
PROCESS_LINE_WIDTH = 0.02 # cm
PROCESS_COLOR = "#787878"
PROCESS_FONT_SIZE = 11 # pt

FONT_FAMILY = "sans"

MAXIMUM_ORIENTATION = 20

E = ElementMaker(namespace="http://www.lysator.liu.se/~alla/dia/",
                 nsmap={"dia" : "http://www.lysator.liu.se/~alla/dia/"})

#########################################################################
#########################################################################

def _attribute(attr_name, attr_value):
    if isinstance(attr_value, etree._Element):
        return attr_value
    elif isinstance(attr_value, DiaAttribute):
        return attr_value.elements()
    else:
        if isinstance(attr_value, types.StringTypes):
            attr_type = "string"
        elif isinstance(attr_value, types.FloatType):
            attr_type = "real"
            attr_value = unicode(attr_value)
        elif isinstance(attr_value, types.BooleanType):
            attr_type = "boolean"
            attr_value = "true" if attr_value else "false"
        elif isinstance(attr_value, types.IntType):
            attr_type = "int"
            attr_value = unicode(attr_value)
        else:
            raise ValueError("unknown attr type %s" % (repr(attr_value)))

        if attr_type == "string":
            return getattr(E, attr_type)(attr_value)
        else:
            return getattr(E, attr_type)(val=attr_value)

def attributes(attrs):
    elements = []
    for attr_name, attr_value in attrs:
        if isinstance(attr_value, (types.ListType, types.TupleType)):
            children = [_attribute(attr_name, value) for value in attr_value]
        else:
            children = [_attribute(attr_name, attr_value)]
        elements.append(E.attribute(*children, name=attr_name))

    return elements

#########################################################################

def text_box_size(txt, font_size):
    """
    @return: approximate width and height
    """
    width = len(txt) * font_size * 0.6 * PT_TO_CM
    height = font_size * PT_TO_CM
    return width, height

#########################################################################
#########################################################################

class DiaAttribute(object):
    def __init__(self, val):
        self.val = val

    def elements(self):
        return getattr(E, self.TYPE)(val=self.val)

##################################################

class ColorAttribute(DiaAttribute):
    TYPE = "color"

##################################################

class PointAttribute(DiaAttribute):
    TYPE = "point"

#########################################################################
#########################################################################

class DiaObject(object):
    pass

##################################################

class LineObject(DiaObject):
    def __init__(self, x1, y1, x2, y2, width, color):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.width = width
        self.color = color

    def more(self):
        return []

    def tree(self, _id):
        attrs = attributes([
            ("obj_pos", PointAttribute("%f,%f" % (self.x1, self.y1))),
            ("conn_endpoints", (
                PointAttribute("%f,%f" % (self.x1, self.y1)),
                PointAttribute("%f,%f" % (self.x2, self.y2))
              )),
            ("numcp", 1),
            ("line_width", self.width),
            ("line_color", ColorAttribute(self.color))
          ]) + self.more()
        return E.object(*attrs, type="Standard - Line", version="0", id=_id)

##################################################

class ArrowObject(LineObject):
    def more(self):
        return attributes([
            ("end_arrow", E.enum(val="1")),
            ("end_arrow_length", ARROW_LENGTH),
            ("end_arrow_width", ARROW_WIDTH)
          ])

##################################################

class TextObject(DiaObject):
    def __init__(self, text, x, y, color, size):
        self.text = text
        self.x = x
        self.y = y
        self.color = color
        self.size = size
    
    def tree(self, _id):
        attrs = attributes([
            ("obj_pos", PointAttribute("%f,%f" % (self.x, self.y))),
            ("text", E.composite(
                *attributes([
                  ("string", "#" + self.text + "#"),
                  ("font", E.font(family=FONT_FAMILY, style="0")),
                  ("height", self.size * PT_TO_CM),
                  ("pos", PointAttribute("%f,%f" % (self.x, self.y))),
                  ("color", ColorAttribute(self.color))
                ]), type="text"))
          ])
        return E.object(*attrs, type="Standard - Text", version="1", id=_id)

##################################################

class PointObject(DiaObject):
    def __init__(self, x, y, diameter, color):
        self.x = x
        self.y = y
        self.diameter = diameter
        self.color = color

    def tree(self, _id):
        attrs = attributes([
            ("obj_pos", PointAttribute("%f,%f" % (self.x, self.y))),
            ("elem_corner", PointAttribute("%f,%f" % (self.x, self.y))),
            ("elem_width", self.diameter),
            ("elem_height", self.diameter),
            ("border_width", 0.0),
            ("inner_color", ColorAttribute(self.color)),
            ("border_color", ColorAttribute(self.color))
          ])
        return E.object(*attrs, type="Standard - Ellipse", version="0", id=_id)

#########################################################################
#########################################################################

class DiaConvertor(object):
    def __init__(self):
        self.input_data = None
        self.dia_document = None
        self.id_counter = 0
        self.time_start = None
        self.time_end = None
        self.process_pos = {}
        self.labels_per_line = {}

    #####################################################
        
    def convert(self):
        self.input_data = yaml.load(sys.stdin)

        times = [packet["src"]["ts"] for packet in self.input_data["traffic"]]
        times += [packet["dst"]["ts"] for packet in self.input_data["traffic"]]
        if "marks" in self.input_data:
            times += [mark["ts"] for mark in self.input_data["marks"]]
        times.sort()
        self.time_start = times[0]
        self.time_end = times[-1]

        self.gen_dia_document()
        print etree.tostring(self.dia_document, pretty_print=True, 
                              xml_declaration=True, encoding="UTF-8")

    #####################################################

    def gen_dia_document(self):
        diagram_data = attributes([
            ("background", ColorAttribute("#ffffff")),
            ("pagebreak", ColorAttribute("#ffffff")),
            ("paper", E.composite(*attributes([
                    ("name", "#A4#"),
                    ("tmargin", 1.0),
                    ("bmargin", 1.0),
                    ("lmargin", 1.0),
                    ("rmargin", 1.0),
                    ("is_portrait", True),
                    ("scaling", 1.0),
                    ("fitto", False)
                  ]), type="paper")),
            ("grid", E.composite(*(attributes([
                    ("width_x", 1.0),
                    ("width_y", 1.0),
                    ("visible_x", 1),
                    ("visible_y", 1),
                  ]) + [E.composite(type="color")]), type="grid")),
            ("color", ColorAttribute("#ffffff")),
            ("guides", E.composite(
                    E.attribute(name="hguides"),
                    E.attribute(name="vguides"),
                    type="guides"))
          ])

        processes = self.draw_processes()
        traffic = self.draw_traffic()
        marks = self.draw_marks()
        labels_connectors, labels = self.draw_traffic_labels()

        layer = processes + labels_connectors + traffic + marks + labels

        self.dia_document = E.diagram(
            E.diagramdata(*diagram_data),
            E.layer(*layer, name="Background", visible="true", active="true")
          )

    #####################################################

    def convert_objects(self, objects):
        return [obj.tree(self.get_id()) for obj in objects]

    #####################################################

    def draw_processes(self):
        dia_processes = []
        for i, process in enumerate(self.input_data["processes"]):
            y = i * PROCESS_SPACING
            self.process_pos[process] = y

            txt_height = PROCESS_FONT_SIZE * PT_TO_CM
            dia_processes.append(TextObject(process + ":", 
                                            0.0, 
                                            y - txt_height / 2.0, 
                                            PROCESS_COLOR,
                                            PROCESS_FONT_SIZE))
            length = self.time_end - self.time_start
            dia_processes.append(LineObject(PROCESS_LINE_VOFFSET, 
                                            y, 
                                            PROCESS_LINE_VOFFSET + length, 
                                            y,
                                            PROCESS_LINE_WIDTH, PROCESS_COLOR))

        return self.convert_objects(dia_processes)

    #####################################################

    def stack_labels(self, labels):
        labels.sort(lambda a,b: cmp(a[1], b[1]))
        for i in range(1, len(labels)):
            label = labels[i]
            while abs(label[4]) < MAXIMUM_ORIENTATION:
                # go backwards and find overlapping label 
                # with the same orientation/level
                for j in range(i - 1, -1, -1):
                    prev_label = labels[j]
                    if prev_label[4] != label[4]:
                        # not on the same level
                        continue
                    width, height = text_box_size(prev_label[0], DATA_FONT_SIZE)
                    prev_end = prev_label[1] + width
                    if label[1] < prev_end:
                        # overlapping found, proceed to next level
                        break
                else:
                    # not found, label is free to stay on its level
                    break # break the "while"

                # try next level
                if label[4] < 0:
                    label[4] -= 1
                else:
                    label[4] += 1
            
    #####################################################

    def draw_traffic(self):
        dia_traffic_arrows = []
        for packet in self.input_data["traffic"]:
            src_process, src_time = packet["src"]["p"], packet["src"]["ts"]
            dst_process, dst_time = packet["dst"]["p"], packet["dst"]["ts"]
            data = packet["data"]
            color = packet.get("color", DATA_COLOR)
            x1 = PROCESS_LINE_VOFFSET + src_time - self.time_start
            y1 = self.process_pos[src_process]
            x2 = PROCESS_LINE_VOFFSET + dst_time - self.time_start
            y2 = self.process_pos[dst_process]

            dia_traffic_arrows.append(ArrowObject(x1, y1, x2, y2,
                                         DATA_LINE_WIDTH, color))

            # put the text on the right side of the process line
            if y1 < y2:
                orientation1 = -1
                orientation2 = 0
            else:
                orientation1 = 0
                orientation2 = -1
            key = "%s%i" % (src_process, orientation1)
            labels = self.labels_per_line.setdefault(key, [])
            labels.append(["S(%s)" % data, x1, y1, color, orientation1])
            key = "%s%i" % (dst_process, orientation2)
            labels = self.labels_per_line.setdefault(key, [])
            labels.append(["R(%s)" % data, x2, y2, color, orientation2])


        return self.convert_objects(dia_traffic_arrows)

    #####################################################

    def draw_marks(self):
        if "marks" not in self.input_data:
            return []

        dia_marks = []
        for mark in self.input_data["marks"]:
            x = PROCESS_LINE_VOFFSET + mark["ts"] - self.time_start
            y = self.process_pos[mark["p"]]
            color = mark.get("color", MARK_COLOR)
            dia_marks.append(PointObject(x - MARK_SIZE / 2.0, 
                                        y - MARK_SIZE / 2.0,
                                        MARK_SIZE, color))
            orientation = -1
            key = "%s%i" % (mark["p"], orientation)
            labels = self.labels_per_line.setdefault(key, [])
            labels.append([mark["text"], x, y, color, orientation])

        return self.convert_objects(dia_marks)

    #####################################################

    def draw_traffic_labels(self):
        dia_traffic_labels = []
        dia_connectors = []

        for labels in self.labels_per_line.values():
            self.stack_labels(labels)
            for data, x, y, color, orientation in labels:
                height = DATA_FONT_SIZE * PT_TO_CM
                ly = y + orientation * height
                lcy = ly + height / 2.0
                dia_connectors.append(LineObject(x, y, x, lcy,
                                            LABEL_CONNECTOR_WIDTH,
                                            LABEL_CONNECTOR_COLOR))
                dia_traffic_labels.append(TextObject(data, x, ly, color,
                                                      DATA_FONT_SIZE))

        return (self.convert_objects(dia_connectors),
               self.convert_objects(dia_traffic_labels))

    #####################################################

    def get_id(self):
        self.id_counter += 1
        return str(self.id_counter)

#########################################################################
#########################################################################

if __name__ == "__main__":
    DiaConvertor().convert()
